@@ -5,27 +5,28 @@
  *	  commands.  At one time acted as an interface between the Lisp and C
  *	  systems.
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/tcop/utility.c,v 1.245 2005/10/15 02:49:27 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/tcop/utility.c,v 1.269.2.1 2009/12/09 21:58:29 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
 #include "access/twophase.h"
+#include "access/xact.h"
 #include "catalog/catalog.h"
 #include "catalog/namespace.h"
+#include "catalog/toasting.h"
 #include "commands/alter.h"
 #include "commands/async.h"
 #include "commands/cluster.h"
 #include "commands/comment.h"
-#include "commands/copy.h"
 #include "commands/conversioncmds.h"
+#include "commands/copy.h"
 #include "commands/dbcommands.h"
 #include "commands/defrem.h"
 #include "commands/explain.h"
@@ -43,9 +44,6 @@
 #include "commands/vacuum.h"
 #include "commands/view.h"
 #include "miscadmin.h"
-#include "nodes/makefuncs.h"
-#include "parser/parse_expr.h"
-#include "parser/parse_type.h"
 #include "postmaster/bgwriter.h"
 #include "rewrite/rewriteDefine.h"
 #include "rewrite/rewriteRemove.h"
@@ -54,7 +52,6 @@
 #include "tcop/utility.h"
 #include "utils/acl.h"
 #include "utils/guc.h"
-#include "utils/lsyscache.h"
 #include "utils/syscache.h"
 
 
@@ -67,6 +64,7 @@ struct msgstrings
 	char		kind;
 	int			nonexistent_code;
 	const char *nonexistent_msg;
+	const char *skipping_msg;
 	const char *nota_msg;
 	const char *drophint_msg;
 };
@@ -75,26 +73,31 @@ static const struct msgstrings msgstringarray[] = {
 	{RELKIND_RELATION,
 		ERRCODE_UNDEFINED_TABLE,
 		gettext_noop("table \"%s\" does not exist"),
+		gettext_noop("table \"%s\" does not exist, skipping"),
 		gettext_noop("\"%s\" is not a table"),
 	gettext_noop("Use DROP TABLE to remove a table.")},
 	{RELKIND_SEQUENCE,
 		ERRCODE_UNDEFINED_TABLE,
 		gettext_noop("sequence \"%s\" does not exist"),
+		gettext_noop("sequence \"%s\" does not exist, skipping"),
 		gettext_noop("\"%s\" is not a sequence"),
 	gettext_noop("Use DROP SEQUENCE to remove a sequence.")},
 	{RELKIND_VIEW,
 		ERRCODE_UNDEFINED_TABLE,
 		gettext_noop("view \"%s\" does not exist"),
+		gettext_noop("view \"%s\" does not exist, skipping"),
 		gettext_noop("\"%s\" is not a view"),
 	gettext_noop("Use DROP VIEW to remove a view.")},
 	{RELKIND_INDEX,
 		ERRCODE_UNDEFINED_OBJECT,
 		gettext_noop("index \"%s\" does not exist"),
+		gettext_noop("index \"%s\" does not exist, skipping"),
 		gettext_noop("\"%s\" is not an index"),
 	gettext_noop("Use DROP INDEX to remove an index.")},
 	{RELKIND_COMPOSITE_TYPE,
 		ERRCODE_UNDEFINED_OBJECT,
 		gettext_noop("type \"%s\" does not exist"),
+		gettext_noop("type \"%s\" does not exist, skipping"),
 		gettext_noop("\"%s\" is not a type"),
 	gettext_noop("Use DROP TYPE to remove a type.")},
 	{'\0', 0, NULL, NULL, NULL}
@@ -132,23 +135,40 @@ DropErrorMsgWrongType(char *relname, char wrongkind, char rightkind)
  * non-existent relation
  */
 static void
-DropErrorMsgNonExistent(RangeVar *rel, char rightkind)
+DropErrorMsgNonExistent(RangeVar *rel, char rightkind, bool missing_ok)
 {
 	const struct msgstrings *rentry;
 
 	for (rentry = msgstringarray; rentry->kind != '\0'; rentry++)
 	{
 		if (rentry->kind == rightkind)
-			ereport(ERROR,
-					(errcode(rentry->nonexistent_code),
-					 errmsg(rentry->nonexistent_msg, rel->relname)));
+		{
+			if (!missing_ok)
+			{
+				ereport(ERROR,
+						(errcode(rentry->nonexistent_code),
+						 errmsg(rentry->nonexistent_msg, rel->relname)));
+			}
+			else
+			{
+				ereport(NOTICE, (errmsg(rentry->skipping_msg, rel->relname)));
+				break;
+			}
+		}
 	}
 
-	Assert(false);				/* Should be impossible */
+	Assert(rentry->kind != '\0');		/* Should be impossible */
 }
 
-static void
-CheckDropPermissions(RangeVar *rel, char rightkind)
+/*
+ * returns false if missing_ok is true and the object does not exist,
+ * true if object exists and permissions are OK,
+ * errors otherwise
+ *
+ */
+
+static bool
+CheckDropPermissions(RangeVar *rel, char rightkind, bool missing_ok)
 {
 	Oid			relOid;
 	HeapTuple	tuple;
@@ -156,7 +176,10 @@ CheckDropPermissions(RangeVar *rel, char rightkind)
 
 	relOid = RangeVarGetRelid(rel, true);
 	if (!OidIsValid(relOid))
-		DropErrorMsgNonExistent(rel, rightkind);
+	{
+		DropErrorMsgNonExistent(rel, rightkind, missing_ok);
+		return false;
+	}
 
 	tuple = SearchSysCache(RELOID,
 						   ObjectIdGetDatum(relOid),
@@ -183,6 +206,8 @@ CheckDropPermissions(RangeVar *rel, char rightkind)
 						rel->relname)));
 
 	ReleaseSysCache(tuple);
+
+	return true;
 }
 
 /*
@@ -305,7 +330,6 @@ check_xact_readonly(Node *parsetree)
 		case T_CreateTrigStmt:
 		case T_CompositeTypeStmt:
 		case T_ViewStmt:
-		case T_RemoveAggrStmt:
 		case T_DropCastStmt:
 		case T_DropStmt:
 		case T_DropdbStmt:
@@ -313,12 +337,13 @@ check_xact_readonly(Node *parsetree)
 		case T_RemoveFuncStmt:
 		case T_DropRoleStmt:
 		case T_DropPLangStmt:
-		case T_RemoveOperStmt:
 		case T_RemoveOpClassStmt:
 		case T_DropPropertyStmt:
 		case T_GrantStmt:
 		case T_GrantRoleStmt:
 		case T_TruncateStmt:
+		case T_DropOwnedStmt:
+		case T_ReassignOwnedStmt:
 			ereport(ERROR,
 					(errcode(ERRCODE_READ_ONLY_SQL_TRANSACTION),
 					 errmsg("transaction is read-only")));
@@ -330,12 +355,31 @@ check_xact_readonly(Node *parsetree)
 }
 
 
+/*
+ * CheckRestrictedOperation: throw error for hazardous command if we're
+ * inside a security restriction context.
+ *
+ * This is needed to protect session-local state for which there is not any
+ * better-defined protection mechanism, such as ownership.
+ */
+static void
+CheckRestrictedOperation(const char *cmdname)
+{
+	if (InSecurityRestrictedOperation())
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 /* translator: %s is name of a SQL command, eg PREPARE */
+				 errmsg("cannot execute %s within security-restricted operation",
+						cmdname)));
+}
+
+
 /*
  * ProcessUtility
  *		general utility function invoker
  *
  *	parsetree: the parse tree for the utility statement
- *	params: parameters to use during execution (currently only used by DECLARE)
+ *	params: parameters to use during execution
  *	dest: where to send results
  *	completionTag: points to a buffer of size COMPLETION_TAG_BUFSIZE
  *		in which to store a command completion status string.
@@ -474,6 +518,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ClosePortalStmt *stmt = (ClosePortalStmt *) parsetree;
 
+				CheckRestrictedOperation("CLOSE");
 				PerformPortalClose(stmt->portalname);
 			}
 			break;
@@ -502,7 +547,7 @@ ProcessUtility(Node *parsetree,
 				 * secondary relation too.
 				 */
 				CommandCounterIncrement();
-				AlterTableCreateToastTable(relOid, true);
+				AlterTableCreateToastTable(relOid);
 			}
 			break;
 
@@ -528,31 +573,36 @@ ProcessUtility(Node *parsetree,
 					{
 						case OBJECT_TABLE:
 							rel = makeRangeVarFromNameList(names);
-							CheckDropPermissions(rel, RELKIND_RELATION);
-							RemoveRelation(rel, stmt->behavior);
+							if (CheckDropPermissions(rel, RELKIND_RELATION,
+													 stmt->missing_ok))
+								RemoveRelation(rel, stmt->behavior);
 							break;
 
 						case OBJECT_SEQUENCE:
 							rel = makeRangeVarFromNameList(names);
-							CheckDropPermissions(rel, RELKIND_SEQUENCE);
-							RemoveRelation(rel, stmt->behavior);
+							if (CheckDropPermissions(rel, RELKIND_SEQUENCE,
+													 stmt->missing_ok))
+								RemoveRelation(rel, stmt->behavior);
 							break;
 
 						case OBJECT_VIEW:
 							rel = makeRangeVarFromNameList(names);
-							CheckDropPermissions(rel, RELKIND_VIEW);
-							RemoveView(rel, stmt->behavior);
+							if (CheckDropPermissions(rel, RELKIND_VIEW,
+													 stmt->missing_ok))
+								RemoveView(rel, stmt->behavior);
 							break;
 
 						case OBJECT_INDEX:
 							rel = makeRangeVarFromNameList(names);
-							CheckDropPermissions(rel, RELKIND_INDEX);
-							RemoveIndex(rel, stmt->behavior);
+							if (CheckDropPermissions(rel, RELKIND_INDEX,
+													 stmt->missing_ok))
+								RemoveIndex(rel, stmt->behavior);
 							break;
 
 						case OBJECT_TYPE:
 							/* RemoveType does its own permissions checks */
-							RemoveType(names, stmt->behavior);
+							RemoveType(names, stmt->behavior,
+									   stmt->missing_ok);
 							break;
 
 						case OBJECT_DOMAIN:
@@ -560,11 +610,13 @@ ProcessUtility(Node *parsetree,
 							/*
 							 * RemoveDomain does its own permissions checks
 							 */
-							RemoveDomain(names, stmt->behavior);
+							RemoveDomain(names, stmt->behavior,
+										 stmt->missing_ok);
 							break;
 
 						case OBJECT_CONVERSION:
-							DropConversionCommand(names, stmt->behavior);
+							DropConversionCommand(names, stmt->behavior,
+												  stmt->missing_ok);
 							break;
 
 						case OBJECT_SCHEMA:
@@ -572,7 +624,8 @@ ProcessUtility(Node *parsetree,
 							/*
 							 * RemoveSchema does its own permissions checks
 							 */
-							RemoveSchema(names, stmt->behavior);
+							RemoveSchema(names, stmt->behavior,
+										 stmt->missing_ok);
 							break;
 
 						default:
@@ -590,11 +643,7 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_TruncateStmt:
-			{
-				TruncateStmt *stmt = (TruncateStmt *) parsetree;
-
-				ExecuteTruncate(stmt->relations);
-			}
+			ExecuteTruncate((TruncateStmt *) parsetree);
 			break;
 
 		case T_CommentStmt:
@@ -602,18 +651,27 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_CopyStmt:
-			DoCopy((CopyStmt *) parsetree);
+			{
+				uint64		processed = DoCopy((CopyStmt *) parsetree);
+
+				if (completionTag)
+					snprintf(completionTag, COMPLETION_TAG_BUFSIZE,
+							 "COPY " UINT64_FORMAT, processed);
+			}
 			break;
 
 		case T_PrepareStmt:
+			CheckRestrictedOperation("PREPARE");
 			PrepareQuery((PrepareStmt *) parsetree);
 			break;
 
 		case T_ExecuteStmt:
-			ExecuteQuery((ExecuteStmt *) parsetree, dest, completionTag);
+			ExecuteQuery((ExecuteStmt *) parsetree, params,
+						 dest, completionTag);
 			break;
 
 		case T_DeallocateStmt:
+			CheckRestrictedOperation("DEALLOCATE");
 			DeallocateQuery((DeallocateStmt *) parsetree);
 			break;
 
@@ -689,8 +747,7 @@ ProcessUtility(Node *parsetree,
 			break;
 
 			/*
-			 * ******************************** object creation / destruction ********************************
-			 *
+			 * **************** object creation / destruction *****************
 			 */
 		case T_DefineStmt:
 			{
@@ -699,12 +756,15 @@ ProcessUtility(Node *parsetree,
 				switch (stmt->kind)
 				{
 					case OBJECT_AGGREGATE:
-						DefineAggregate(stmt->defnames, stmt->definition);
+						DefineAggregate(stmt->defnames, stmt->args,
+										stmt->oldstyle, stmt->definition);
 						break;
 					case OBJECT_OPERATOR:
+						Assert(stmt->args == NIL);
 						DefineOperator(stmt->defnames, stmt->definition);
 						break;
 					case OBJECT_TYPE:
+						Assert(stmt->args == NIL);
 						DefineType(stmt->defnames, stmt->definition);
 						break;
 					default:
@@ -743,6 +803,9 @@ ProcessUtility(Node *parsetree,
 			{
 				IndexStmt  *stmt = (IndexStmt *) parsetree;
 
+				if (stmt->concurrent)
+					PreventTransactionChain(stmt, "CREATE INDEX CONCURRENTLY");
+
 				CheckRelationOwnership(stmt->relation, true);
 
 				DefineIndex(stmt->relation,		/* relation */
@@ -753,13 +816,15 @@ ProcessUtility(Node *parsetree,
 							stmt->indexParams,	/* parameters */
 							(Expr *) stmt->whereClause,
 							stmt->rangetable,
+							stmt->options,
 							stmt->unique,
 							stmt->primary,
 							stmt->isconstraint,
 							false,		/* is_alter_table */
 							true,		/* check_rights */
 							false,		/* skip_build */
-							false);		/* quiet */
+							false,		/* quiet */
+							stmt->concurrent);	/* concurrent */
 			}
 			break;
 
@@ -775,16 +840,27 @@ ProcessUtility(Node *parsetree,
 			AlterSequence((AlterSeqStmt *) parsetree);
 			break;
 
-		case T_RemoveAggrStmt:
-			RemoveAggregate((RemoveAggrStmt *) parsetree);
-			break;
-
 		case T_RemoveFuncStmt:
-			RemoveFunction((RemoveFuncStmt *) parsetree);
-			break;
+			{
+				RemoveFuncStmt *stmt = (RemoveFuncStmt *) parsetree;
 
-		case T_RemoveOperStmt:
-			RemoveOperator((RemoveOperStmt *) parsetree);
+				switch (stmt->kind)
+				{
+					case OBJECT_FUNCTION:
+						RemoveFunction(stmt);
+						break;
+					case OBJECT_AGGREGATE:
+						RemoveAggregate(stmt);
+						break;
+					case OBJECT_OPERATOR:
+						RemoveOperator(stmt);
+						break;
+					default:
+						elog(ERROR, "unrecognized object type: %d",
+							 (int) stmt->kind);
+						break;
+				}
+			}
 			break;
 
 		case T_CreatedbStmt:
@@ -803,7 +879,7 @@ ProcessUtility(Node *parsetree,
 			{
 				DropdbStmt *stmt = (DropdbStmt *) parsetree;
 
-				dropdb(stmt->dbname);
+				dropdb(stmt->dbname, stmt->missing_ok);
 			}
 			break;
 
@@ -820,6 +896,7 @@ ProcessUtility(Node *parsetree,
 			{
 				ListenStmt *stmt = (ListenStmt *) parsetree;
 
+				CheckRestrictedOperation("LISTEN");
 				Async_Listen(stmt->relation->relname);
 			}
 			break;
@@ -828,6 +905,7 @@ ProcessUtility(Node *parsetree,
 			{
 				UnlistenStmt *stmt = (UnlistenStmt *) parsetree;
 
+				CheckRestrictedOperation("UNLISTEN");
 				Async_Unlisten(stmt->relation->relname);
 			}
 			break;
@@ -836,12 +914,9 @@ ProcessUtility(Node *parsetree,
 			{
 				LoadStmt   *stmt = (LoadStmt *) parsetree;
 
-				if (!superuser())
-					ereport(ERROR,
-							(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
-							 errmsg("must be superuser to do LOAD")));
 				closeAllVfds(); /* probably not necessary... */
-				load_file(stmt->filename);
+				/* Allowed names are restricted if you're not superuser */
+				load_file(stmt->filename, !superuser());
 			}
 			break;
 
@@ -854,7 +929,7 @@ ProcessUtility(Node *parsetree,
 			break;
 
 		case T_ExplainStmt:
-			ExplainQuery((ExplainStmt *) parsetree, dest);
+			ExplainQuery((ExplainStmt *) parsetree, params, dest);
 			break;
 
 		case T_VariableSetStmt:
@@ -934,12 +1009,12 @@ ProcessUtility(Node *parsetree,
 					case OBJECT_RULE:
 						/* RemoveRewriteRule checks permissions */
 						RemoveRewriteRule(relId, stmt->property,
-										  stmt->behavior);
+										  stmt->behavior, stmt->missing_ok);
 						break;
 					case OBJECT_TRIGGER:
 						/* DropTrigger checks permissions */
 						DropTrigger(relId, stmt->property,
-									stmt->behavior);
+									stmt->behavior, stmt->missing_ok);
 						break;
 					default:
 						elog(ERROR, "unrecognized object type: %d",
@@ -983,6 +1058,14 @@ ProcessUtility(Node *parsetree,
 			DropRole((DropRoleStmt *) parsetree);
 			break;
 
+		case T_DropOwnedStmt:
+			DropOwnedObjects((DropOwnedStmt *) parsetree);
+			break;
+
+		case T_ReassignOwnedStmt:
+			ReassignOwnedObjects((ReassignOwnedStmt *) parsetree);
+			break;
+
 		case T_LockStmt:
 			LockTableCommand((LockStmt *) parsetree);
 			break;
@@ -1090,7 +1173,7 @@ UtilityReturnsTuples(Node *parsetree)
 				switch (ChoosePortalStrategy(entry->query_list))
 				{
 					case PORTAL_ONE_SELECT:
-						return true;
+					case PORTAL_ONE_RETURNING:
 					case PORTAL_UTIL_SELECT:
 						return true;
 					case PORTAL_MULTI_QUERY:
@@ -1166,6 +1249,38 @@ UtilityTupleDescriptor(Node *parsetree)
 }
 
 
+/*
+ * QueryReturnsTuples
+ *		Return "true" if this Query will send output to the destination.
+ */
+bool
+QueryReturnsTuples(Query *parsetree)
+{
+	switch (parsetree->commandType)
+	{
+		case CMD_SELECT:
+			/* returns tuples ... unless it's SELECT INTO */
+			if (parsetree->into == NULL)
+				return true;
+			break;
+		case CMD_INSERT:
+		case CMD_UPDATE:
+		case CMD_DELETE:
+			/* the forms with RETURNING return tuples */
+			if (parsetree->returningList)
+				return true;
+			break;
+		case CMD_UTILITY:
+			return UtilityReturnsTuples(parsetree->utilityStmt);
+		case CMD_UNKNOWN:
+		case CMD_NOTHING:
+			/* probably shouldn't get here */
+			break;
+	}
+	return false;				/* default */
+}
+
+
 /*
  * CreateCommandTag
  *		utility to get a string representation of the
@@ -1528,16 +1643,21 @@ CreateCommandTag(Node *parsetree)
 			tag = "ALTER SEQUENCE";
 			break;
 
-		case T_RemoveAggrStmt:
-			tag = "DROP AGGREGATE";
-			break;
-
 		case T_RemoveFuncStmt:
-			tag = "DROP FUNCTION";
-			break;
-
-		case T_RemoveOperStmt:
-			tag = "DROP OPERATOR";
+			switch (((RemoveFuncStmt *) parsetree)->kind)
+			{
+				case OBJECT_FUNCTION:
+					tag = "DROP FUNCTION";
+					break;
+				case OBJECT_AGGREGATE:
+					tag = "DROP AGGREGATE";
+					break;
+				case OBJECT_OPERATOR:
+					tag = "DROP OPERATOR";
+					break;
+				default:
+					tag = "???";
+			}
 			break;
 
 		case T_CreatedbStmt:
@@ -1641,6 +1761,14 @@ CreateCommandTag(Node *parsetree)
 			tag = "DROP ROLE";
 			break;
 
+		case T_DropOwnedStmt:
+			tag = "DROP OWNED";
+			break;
+
+		case T_ReassignOwnedStmt:
+			tag = "REASSIGN OWNED";
+			break;
+
 		case T_LockStmt:
 			tag = "LOCK TABLE";
 			break;
@@ -1711,6 +1839,8 @@ CreateQueryTag(Query *parsetree)
 {
 	const char *tag;
 
+	Assert(IsA(parsetree, Query));
+
 	switch (parsetree->commandType)
 	{
 		case CMD_SELECT:
@@ -1723,7 +1853,7 @@ CreateQueryTag(Query *parsetree)
 				tag = "SELECT INTO";
 			else if (parsetree->rowMarks != NIL)
 			{
-				if (parsetree->forUpdate)
+				if (((RowMarkClause *) linitial(parsetree->rowMarks))->forUpdate)
 					tag = "SELECT FOR UPDATE";
 				else
 					tag = "SELECT FOR SHARE";
@@ -1752,3 +1882,394 @@ CreateQueryTag(Query *parsetree)
 
 	return tag;
 }
+
+
+/*
+ * GetCommandLogLevel
+ *		utility to get the minimum log_statement level for a command,
+ *		given a raw (un-analyzed) parsetree.
+ *
+ * This must handle all raw command types, but since the vast majority
+ * of 'em are utility commands, it seems sensible to keep it here.
+ */
+LogStmtLevel
+GetCommandLogLevel(Node *parsetree)
+{
+	LogStmtLevel lev;
+
+	switch (nodeTag(parsetree))
+	{
+		case T_InsertStmt:
+		case T_DeleteStmt:
+		case T_UpdateStmt:
+			lev = LOGSTMT_MOD;
+			break;
+
+		case T_SelectStmt:
+			if (((SelectStmt *) parsetree)->into)
+				lev = LOGSTMT_DDL;		/* CREATE AS, SELECT INTO */
+			else
+				lev = LOGSTMT_ALL;
+			break;
+
+		case T_TransactionStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_DeclareCursorStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ClosePortalStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_FetchStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_CreateDomainStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateSchemaStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateTableSpaceStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropTableSpaceStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_TruncateStmt:
+			lev = LOGSTMT_MOD;
+			break;
+
+		case T_CommentStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CopyStmt:
+			if (((CopyStmt *) parsetree)->is_from)
+				lev = LOGSTMT_MOD;
+			else
+				lev = LOGSTMT_ALL;
+			break;
+
+		case T_RenameStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterObjectSchemaStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterOwnerStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterTableStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterDomainStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterFunctionStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_GrantStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_GrantRoleStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DefineStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CompositeTypeStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_ViewStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateFunctionStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_IndexStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_RuleStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateSeqStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterSeqStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_RemoveFuncStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreatedbStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterDatabaseStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterDatabaseSetStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropdbStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_NotifyStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ListenStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_UnlistenStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_LoadStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ClusterStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_VacuumStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ExplainStmt:
+			{
+				ExplainStmt *stmt = (ExplainStmt *) parsetree;
+
+				/* Look through an EXPLAIN ANALYZE to the contained stmt */
+				if (stmt->analyze)
+					return GetCommandLogLevel((Node *) stmt->query);
+				/* Plain EXPLAIN isn't so interesting */
+				lev = LOGSTMT_ALL;
+			}
+			break;
+
+		case T_VariableSetStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_VariableShowStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_VariableResetStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_CreateTrigStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropPropertyStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreatePLangStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropPLangStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateRoleStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterRoleStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_AlterRoleSetStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropRoleStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropOwnedStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_ReassignOwnedStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_LockStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ConstraintsSetStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_CheckPointStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_ReindexStmt:
+			lev = LOGSTMT_ALL;	/* should this be DDL? */
+			break;
+
+		case T_CreateConversionStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateCastStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_DropCastStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_CreateOpClassStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_RemoveOpClassStmt:
+			lev = LOGSTMT_DDL;
+			break;
+
+		case T_PrepareStmt:
+			{
+				PrepareStmt *stmt = (PrepareStmt *) parsetree;
+
+				/* Look through a PREPARE to the contained stmt */
+				return GetCommandLogLevel((Node *) stmt->query);
+			}
+			break;
+
+		case T_ExecuteStmt:
+			{
+				ExecuteStmt *stmt = (ExecuteStmt *) parsetree;
+				PreparedStatement *pstmt;
+				ListCell   *l;
+
+				/* Look through an EXECUTE to the referenced stmt(s) */
+				lev = LOGSTMT_ALL;
+				pstmt = FetchPreparedStatement(stmt->name, false);
+				if (pstmt)
+				{
+					foreach(l, pstmt->query_list)
+					{
+						Query	   *query = (Query *) lfirst(l);
+						LogStmtLevel stmt_lev;
+
+						stmt_lev = GetQueryLogLevel(query);
+						lev = Min(lev, stmt_lev);
+					}
+				}
+			}
+			break;
+
+		case T_DeallocateStmt:
+			lev = LOGSTMT_ALL;
+			break;
+
+		case T_Query:
+
+			/*
+			 * In complicated situations (eg, EXPLAIN ANALYZE in an extended
+			 * Query protocol), we might find an already-analyzed query within
+			 * a utility statement.  Cope.
+			 */
+			lev = GetQueryLogLevel((Query *) parsetree);
+			break;
+
+		default:
+			elog(WARNING, "unrecognized node type: %d",
+				 (int) nodeTag(parsetree));
+			lev = LOGSTMT_ALL;
+			break;
+	}
+
+	return lev;
+}
+
+/*
+ * GetQueryLogLevel
+ *		utility to get the minimum log_statement level for a Query operation.
+ *
+ * This is exactly like GetCommandLogLevel, except it works on a Query
+ * that has already been through parse analysis (and possibly further).
+ */
+LogStmtLevel
+GetQueryLogLevel(Query *parsetree)
+{
+	LogStmtLevel lev;
+
+	Assert(IsA(parsetree, Query));
+
+	switch (parsetree->commandType)
+	{
+		case CMD_SELECT:
+			if (parsetree->into != NULL)
+				lev = LOGSTMT_DDL;		/* CREATE AS, SELECT INTO */
+			else
+				lev = LOGSTMT_ALL;
+			break;
+
+		case CMD_UPDATE:
+		case CMD_INSERT:
+		case CMD_DELETE:
+			lev = LOGSTMT_MOD;
+			break;
+
+		case CMD_UTILITY:
+			lev = GetCommandLogLevel(parsetree->utilityStmt);
+			break;
+
+		default:
+			elog(WARNING, "unrecognized commandType: %d",
+				 (int) parsetree->commandType);
+			lev = LOGSTMT_ALL;
+			break;
+	}
+
+	return lev;
+}
