@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/adt/datetime.c,v 1.174.2.4 2008/06/09 19:34:16 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/adt/datetime.c,v 1.174.2.7 2009/08/18 21:23:35 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -840,7 +840,8 @@ DecodeDateTime(char **field, int *ftype, int nf,
 				 */
 				/* test for > 24:00:00 */
 				if (tm->tm_hour > 24 ||
-					(tm->tm_hour == 24 && (tm->tm_min > 0 || tm->tm_sec > 0)))
+					(tm->tm_hour == 24 &&
+					 (tm->tm_min > 0 || tm->tm_sec > 0 || *fsec > 0)))
 					return DTERR_FIELD_OVERFLOW;
 				break;
 
@@ -1957,14 +1958,12 @@ DecodeTimeOnly(char **field, int *ftype, int nf,
 	if (tm->tm_hour < 0 || tm->tm_min < 0 || tm->tm_min > 59 ||
 		tm->tm_sec < 0 || tm->tm_sec > 60 || tm->tm_hour > 24 ||
 		/* test for > 24:00:00 */
+		(tm->tm_hour == 24 &&
+		 (tm->tm_min > 0 || tm->tm_sec > 0 || *fsec > 0)) ||
 #ifdef HAVE_INT64_TIMESTAMP
-		(tm->tm_hour == 24 && (tm->tm_min > 0 || tm->tm_sec > 0 ||
-							   *fsec > INT64CONST(0))) ||
-		*fsec < INT64CONST(0) || *fsec >= USECS_PER_SEC
+		*fsec < INT64CONST(0) || *fsec > USECS_PER_SEC
 #else
-		(tm->tm_hour == 24 && (tm->tm_min > 0 || tm->tm_sec > 0 ||
-							   *fsec > 0)) ||
-		*fsec < 0 || *fsec >= 1
+		*fsec < 0 || *fsec > 1
 #endif
 		)
 		return DTERR_FIELD_OVERFLOW;
@@ -2269,11 +2268,11 @@ DecodeTime(char *str, int fmask, int *tmask, struct pg_tm * tm, fsec_t *fsec)
 #ifdef HAVE_INT64_TIMESTAMP
 	if (tm->tm_hour < 0 || tm->tm_min < 0 || tm->tm_min > 59 ||
 		tm->tm_sec < 0 || tm->tm_sec > 60 || *fsec < INT64CONST(0) ||
-		*fsec >= USECS_PER_SEC)
+		*fsec > USECS_PER_SEC)
 		return DTERR_FIELD_OVERFLOW;
 #else
 	if (tm->tm_hour < 0 || tm->tm_min < 0 || tm->tm_min > 59 ||
-		tm->tm_sec < 0 || tm->tm_sec > 60 || *fsec < 0 || *fsec >= 1)
+		tm->tm_sec < 0 || tm->tm_sec > 60 || *fsec < 0 || *fsec > 1)
 		return DTERR_FIELD_OVERFLOW;
 #endif
 
@@ -2826,6 +2825,9 @@ DecodeInterval(char **field, int *ftype, int nf, int *dtype, struct pg_tm * tm,
 						break;
 
 					case DTK_MILLISEC:
+						/* avoid overflowing the fsec field */
+						tm->tm_sec += val / 1000;
+						val -= (val / 1000) * 1000;
 #ifdef HAVE_INT64_TIMESTAMP
 						*fsec += (val + fval) * 1000;
 #else
@@ -3711,7 +3713,7 @@ EncodeInterval(struct pg_tm * tm, fsec_t fsec, int style, char *str)
 					is_before = TRUE;
 				}
 				sprintf(cp, "%s%d.%02d secs", is_nonzero ? " " : "",
-						tm->tm_sec, ((int) sec) / 10000);
+						tm->tm_sec, abs((int) rint(sec / 10000.0)));
 				cp += strlen(cp);
 #else
 				fsec += tm->tm_sec;
