@@ -5,9 +5,9 @@
  * to contain some useful information. Mechanism differs wildly across
  * platforms.
  *
- * $PostgreSQL: pgsql/src/backend/utils/misc/ps_status.c,v 1.23 2005/01/01 05:43:08 momjian Exp $
+ * $PostgreSQL: pgsql/src/backend/utils/misc/ps_status.c,v 1.33.2.2 2010/05/27 19:19:58 tgl Exp $
  *
- * Copyright (c) 2000-2005, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2006, PostgreSQL Global Development Group
  * various details abducted from various places
  *--------------------------------------------------------------------
  */
@@ -31,9 +31,10 @@
 #include "utils/ps_status.h"
 
 extern char **environ;
+bool		update_process_title = true;
 
 
-/*------
+/*
  * Alternative ways of updating ps display:
  *
  * PS_USE_SETPROCTITLE
@@ -50,7 +51,9 @@ extern char **environ;
  *	   (some other BSD systems)
  * PS_USE_CLOBBER_ARGV
  *	   write over the argv and environment area
- *	   (most SysV-like systems)
+ *	   (Linux and most SysV-like systems)
+ * PS_USE_WIN32
+ *	   push the string out as the name of a Windows event
  * PS_USE_NONE
  *	   don't update ps display
  *	   (This is the default, as it is safest.)
@@ -63,9 +66,9 @@ extern char **environ;
 #define PS_USE_PS_STRINGS
 #elif (defined(BSD) || defined(__bsdi__) || defined(__hurd__)) && !defined(__darwin__)
 #define PS_USE_CHANGE_ARGV
-#elif defined(__linux__) || defined(_AIX) || defined(__sgi) || (defined(sun) && !defined(BSD)) || defined(ultrix) || defined(__ksr__) || defined(__osf__) || defined(__QNX__) || defined(__svr4__) || defined(__svr5__) || defined(__darwin__)
+#elif defined(__linux__) || defined(_AIX) || defined(__sgi) || (defined(sun) && !defined(BSD)) || defined(ultrix) || defined(__ksr__) || defined(__osf__) || defined(__svr4__) || defined(__svr5__) || defined(__darwin__)
 #define PS_USE_CLOBBER_ARGV
-#elif defined (WIN32)
+#elif defined(WIN32)
 #define PS_USE_WIN32
 #else
 #define PS_USE_NONE
@@ -73,7 +76,7 @@ extern char **environ;
 
 
 /* Different systems want the buffer padded differently */
-#if defined(_AIX) || defined(__linux__) || defined(__QNX__) || defined(__svr4__)
+#if defined(_AIX) || defined(__linux__) || defined(__svr4__)
 #define PS_PADDING '\0'
 #else
 #define PS_PADDING ' '
@@ -81,44 +84,24 @@ extern char **environ;
 
 
 #ifndef PS_USE_CLOBBER_ARGV
-/* all but one options need a buffer to write their ps line in */
+/* all but one option need a buffer to write their ps line in */
 #define PS_BUFFER_SIZE 256
 static char ps_buffer[PS_BUFFER_SIZE];
 static const size_t ps_buffer_size = PS_BUFFER_SIZE;
-
 #else							/* PS_USE_CLOBBER_ARGV */
 static char *ps_buffer;			/* will point to argv area */
 static size_t ps_buffer_size;	/* space determined at run time */
+static size_t last_status_len;	/* use to minimize length of clobber */
 #endif   /* PS_USE_CLOBBER_ARGV */
 
+static size_t ps_buffer_cur_len;		/* nominal strlen(ps_buffer) */
+
 static size_t ps_buffer_fixed_size;		/* size of the constant prefix */
 
 /* save the original argv[] location here */
 static int	save_argc;
 static char **save_argv;
 
-#ifdef WIN32
-	/*
-	 * Win32 does not support showing any changed arguments. To make it
-	 * at all possible to track which backend is doing what, we create
-	 * a named object that can be viewed with for example Process Explorer
-	 */
-static HANDLE ident_handle = INVALID_HANDLE_VALUE;
-static void pgwin32_update_ident(char *ident)
-{
-	char name[PS_BUFFER_SIZE+32];
-
-	if (ident_handle != INVALID_HANDLE_VALUE)
-		CloseHandle(ident_handle);
-
-	sprintf(name,"pgident: %s",ident);
-
-	ident_handle = CreateEvent(NULL,
-							   TRUE,
-							   FALSE,
-							   name);
-}
-#endif
 
 /*
  * Call this early in startup to save the original argc/argv values.
@@ -139,8 +122,8 @@ save_ps_display_args(int argc, char **argv)
 #if defined(PS_USE_CLOBBER_ARGV)
 
 	/*
-	 * If we're going to overwrite the argv area, count the available
-	 * space.  Also move the environment to make additional room.
+	 * If we're going to overwrite the argv area, count the available space.
+	 * Also move the environment to make additional room.
 	 */
 	{
 		char	   *end_of_area = NULL;
@@ -173,7 +156,7 @@ save_ps_display_args(int argc, char **argv)
 		}
 
 		ps_buffer = argv[0];
-		ps_buffer_size = end_of_area - argv[0];
+		last_status_len = ps_buffer_size = end_of_area - argv[0];
 
 		/*
 		 * move the environment out of the way
@@ -193,12 +176,12 @@ save_ps_display_args(int argc, char **argv)
 	 * argument parsing purposes.
 	 *
 	 * (NB: do NOT think to remove the copying of argv[], even though
-	 * postmaster.c finishes looking at argv[] long before we ever
-	 * consider changing the ps display.  On some platforms, getopt()
-	 * keeps pointers into the argv array, and will get horribly confused
-	 * when it is re-called to analyze a subprocess' argument string if
-	 * the argv storage has been clobbered meanwhile.  Other platforms
-	 * have other dependencies on argv[].
+	 * postmaster.c finishes looking at argv[] long before we ever consider
+	 * changing the ps display.  On some platforms, getopt() keeps pointers
+	 * into the argv array, and will get horribly confused when it is
+	 * re-called to analyze a subprocess' argument string if the argv storage
+	 * has been clobbered meanwhile.  Other platforms have other dependencies
+	 * on argv[].
 	 */
 	{
 		char	  **new_argv;
@@ -220,8 +203,7 @@ save_ps_display_args(int argc, char **argv)
 
 		argv = new_argv;
 	}
-#endif   /* PS_USE_CHANGE_ARGV or
-								 * PS_USE_CLOBBER_ARGV */
+#endif   /* PS_USE_CHANGE_ARGV or PS_USE_CLOBBER_ARGV */
 
 	return argv;
 }
@@ -232,7 +214,7 @@ save_ps_display_args(int argc, char **argv)
  */
 void
 init_ps_display(const char *username, const char *dbname,
-				const char *host_info)
+				const char *host_info, const char *initial_str)
 {
 	Assert(username);
 	Assert(dbname);
@@ -246,6 +228,7 @@ init_ps_display(const char *username, const char *dbname,
 	/* no ps display if you didn't call save_ps_display_args() */
 	if (!save_argv)
 		return;
+
 #ifdef PS_USE_CLOBBER_ARGV
 	/* If ps_buffer is a pointer, it might still be null */
 	if (!ps_buffer)
@@ -278,8 +261,8 @@ init_ps_display(const char *username, const char *dbname,
 #ifdef PS_USE_SETPROCTITLE
 
 	/*
-	 * apparently setproctitle() already adds a `progname:' prefix to the
-	 * ps line
+	 * apparently setproctitle() already adds a `progname:' prefix to the ps
+	 * line
 	 */
 	snprintf(ps_buffer, ps_buffer_size,
 			 "%s %s %s ",
@@ -290,12 +273,9 @@ init_ps_display(const char *username, const char *dbname,
 			 username, dbname, host_info);
 #endif
 
-	ps_buffer_fixed_size = strlen(ps_buffer);
-
-#ifdef WIN32
-	pgwin32_update_ident(ps_buffer);
-#endif
+	ps_buffer_cur_len = ps_buffer_fixed_size = strlen(ps_buffer);
 
+	set_ps_display(initial_str, true);
 #endif   /* not PS_USE_NONE */
 }
 
@@ -306,13 +286,13 @@ init_ps_display(const char *username, const char *dbname,
  * indication of what you're currently doing passed in the argument.
  */
 void
-set_ps_display(const char *activity)
+set_ps_display(const char *activity, bool force)
 {
-	/* save tag for possible use by elog.c */
-	if (MyProcPort)
-		MyProcPort->commandTag = activity;
-
 #ifndef PS_USE_NONE
+	/* update_process_title=off disables updates, unless force = true */
+	if (!force && !update_process_title)
+		return;
+
 	/* no ps display for stand-alone backend */
 	if (!IsUnderPostmaster)
 		return;
@@ -324,8 +304,9 @@ set_ps_display(const char *activity)
 #endif
 
 	/* Update ps_buffer to contain both fixed part and activity */
-	StrNCpy(ps_buffer + ps_buffer_fixed_size, activity,
+	strlcpy(ps_buffer + ps_buffer_fixed_size, activity,
 			ps_buffer_size - ps_buffer_fixed_size);
+	ps_buffer_cur_len = strlen(ps_buffer);
 
 	/* Transmit new setting to kernel, if necessary */
 
@@ -338,7 +319,7 @@ set_ps_display(const char *activity)
 		union pstun pst;
 
 		pst.pst_command = ps_buffer;
-		pstat(PSTAT_SETCMD, pst, strlen(ps_buffer), 0, 0);
+		pstat(PSTAT_SETCMD, pst, ps_buffer_cur_len, 0, 0);
 	}
 #endif   /* PS_USE_PSTAT */
 
@@ -348,35 +329,54 @@ set_ps_display(const char *activity)
 #endif   /* PS_USE_PS_STRINGS */
 
 #ifdef PS_USE_CLOBBER_ARGV
-	{
-		int			buflen;
-
-		/* pad unused memory */
-		buflen = strlen(ps_buffer);
-		MemSet(ps_buffer + buflen, PS_PADDING, ps_buffer_size - buflen);
-	}
+	/* pad unused memory; need only clobber remainder of old status string */
+	if (last_status_len > ps_buffer_cur_len)
+		MemSet(ps_buffer + ps_buffer_cur_len, PS_PADDING,
+			   last_status_len - ps_buffer_cur_len);
+	last_status_len = ps_buffer_cur_len;
 #endif   /* PS_USE_CLOBBER_ARGV */
 
-#ifdef WIN32
-	pgwin32_update_ident(ps_buffer);
-#endif
+#ifdef PS_USE_WIN32
+	{
+		/*
+		 * Win32 does not support showing any changed arguments. To make it at
+		 * all possible to track which backend is doing what, we create a
+		 * named object that can be viewed with for example Process Explorer.
+		 */
+		static HANDLE ident_handle = INVALID_HANDLE_VALUE;
+		char		name[PS_BUFFER_SIZE + 32];
 
+		if (ident_handle != INVALID_HANDLE_VALUE)
+			CloseHandle(ident_handle);
+
+		sprintf(name, "pgident(%d): %s", MyProcPid, ps_buffer);
+
+		ident_handle = CreateEvent(NULL, TRUE, FALSE, name);
+	}
+#endif   /* PS_USE_WIN32 */
 #endif   /* not PS_USE_NONE */
 }
 
 
 /*
  * Returns what's currently in the ps display, in case someone needs
- * it.	Note that only the activity part is returned.
+ * it.	Note that only the activity part is returned.  On some platforms
+ * the string will not be null-terminated, so return the effective
+ * length into *displen.
  */
 const char *
-get_ps_display(void)
+get_ps_display(int *displen)
 {
 #ifdef PS_USE_CLOBBER_ARGV
 	/* If ps_buffer is a pointer, it might still be null */
 	if (!ps_buffer)
+	{
+		*displen = 0;
 		return "";
+	}
 #endif
 
+	*displen = (int) (ps_buffer_cur_len - ps_buffer_fixed_size);
+
 	return ps_buffer + ps_buffer_fixed_size;
 }
